IMPLEMENTATION MODULE SimLib0;
 
  (**************************************************************)
  (*                                                            *)
  (*  Module:   SimLib0, V 0.1					*)
  (*                                                            *)
  (*  Programming: Markus Ulrich				*)
  (*                                                            *)
  (**************************************************************)
  (*   LAST MODIFICATION: 15.4.86 				*)
  (**************************************************************)

FROM DMLanguage IMPORT Language, SetLanguage, CurrentLanguage;
FROM DMAlerts 	IMPORT	WriteMessage, ShowAlert;

TYPE  ErrorTypes = (procedureNotImplemented,methodNotImplemented,
		    noModelInstalled, badModelOrder);

VAR h            : REAL; (*integration step*)
    currentMethod: IntegrationMethod;
    CurrentEquationProcedure: EquationProcedure;
    errorCount: CARDINAL; (*counts calls of "Error(NoModelInstalled)"*)



PROCEDURE NoProcedureP;
BEGIN
  CASE CurrentLanguage() OF
    English: 
    
WriteMessage(2,3,' Warning of Library Module "SimLib0":');
WriteMessage(4,3,"       A procedure which is not yet implemented was called.");  

  | German : ;
    
WriteMessage(2,3,' Warnung des Bibliothek-Moduls "SimLib0":');
WriteMessage(4,3,"        Eine noch nicht implementierte Prozedur wurde");  
WriteMessage(5,3,"        aufgerufen.");  

  END; (*CASE*)
END NoProcedureP;


PROCEDURE NoMethodP;
BEGIN
  CASE CurrentLanguage() OF
    English: 
    
WriteMessage(2,3,' Error in procedure "SetIntegrationMethod" ');
WriteMessage(3,3,' of Library Module "SimLib0":');
WriteMessage(5,3,"       The selected method is not yet implemented.");  

  | German : ;
    
WriteMessage(2,3,' Fehler in Prozedur "SetIntegrationMethod"');
WriteMessage(3,3,' des Bibliothek-Moduls "SimLib0":');
WriteMessage(5,3,"        Die gewhlte Methode ist noch nicht implementiert.");  

  END; (*CASE*)
END NoMethodP;


PROCEDURE NoModelP;
BEGIN
  CASE CurrentLanguage() OF
    English: 
    
WriteMessage(2,3,' Error in Library Module "SimLib0":');
WriteMessage(4,3,'       Procedure "Integrate" was called before a model was');  
WriteMessage(5,3,'       installed with "InstallModel".');  
WriteMessage(6,3,'       All variables of the model are initialized to zero!');  

  | German : ;
    
WriteMessage(2,3,' Fehler des Bibliothek-Moduls "SimLib0":');
WriteMessage(4,3,'       Es wurde die Prozedur "Integrate" gerufen, ohne  ');  
WriteMessage(5,3,'       zuvor mit "InstallModel" ein Modell zu installieren.');  
WriteMessage(5,3,'       Alle Modellvariablen werden auf 0.0 initialisiert!');  

  END; (*CASE*)
END NoModelP;


PROCEDURE BadModOrdP;
BEGIN
  CASE CurrentLanguage() OF
    English: 
    
WriteMessage(2,3,' Eror of Library Module "SimLib0":');
WriteMessage(4,3,"       The order of the installed model exceeds nmax.");  

  | German : ;
    
WriteMessage(2,3,' Fehler des Bibliothek-Moduls "SimLib0":');
WriteMessage(4,3,"        Die Ordnung des installierten Modells ist");  
WriteMessage(5,3,"        grsser als nmax.");  

  END; (*CASE*)
END BadModOrdP;


PROCEDURE Error(e: ErrorTypes);
BEGIN (*Error*)
  CASE e OF
    procedureNotImplemented	:   ShowAlert(8,62, NoProcedureP);
  | methodNotImplemented	:   ShowAlert(8,62, NoMethodP);
  | noModelInstalled		:   ShowAlert(8,62, NoModelP);
  | badModelOrder		:   ShowAlert(8,62, BadModOrdP);
  END; (*CASE*)
END Error;
  
  

PROCEDURE InstallModel(m:EquationProcedure);
BEGIN
  CurrentEquationProcedure:= m;
END InstallModel;

	 
PROCEDURE EulerProc(VAR x, xnew: ARRAY OF REAL);
  VAR i,n: CARDINAL;
      dx : ARRAY [0..nmax-1] OF REAL;
BEGIN (*EulerProc*)
  n:= HIGH(x);
  IF n>nmax-1 THEN (*too big model order*) 
    Error(badModelOrder)
  END;
  CurrentEquationProcedure(x,dx);
  FOR i:= 0 TO n DO xnew[i]:= x[i]+h*dx[i] END;
  time:= time+h;
END EulerProc;


PROCEDURE HeunProc(VAR x, xnew: ARRAY OF REAL);
  VAR i,n: CARDINAL;
      dx, dxx : ARRAY [0..nmax-1] OF REAL;
BEGIN (*HeunProc*)
  n:= HIGH(x);
  IF n>nmax-1 THEN (*too big model order*) 
    Error(badModelOrder)
  END;
  CurrentEquationProcedure(x,dx);
  FOR i:= 0 TO n DO xnew[i]:= x[i]+h*dx[i] END;
  CurrentEquationProcedure(xnew,dxx);
  FOR i:= 0 TO n DO xnew[i]:= x[i]+h*(dx[i]+dxx[i])/2.0 END;
  time:= time+h;
END HeunProc;


PROCEDURE RungeKutta4Proc(VAR x, xnew: ARRAY OF REAL);
  (*reference: "Numerische Methoden der Automatik", p. III/7, D.Rufer *)
  VAR i,n: CARDINAL;
      hh: REAL; (*half of integration step h*)
      x1, x2, x3, dx, dx1, dx2, dx3 : ARRAY [0..nmax-1] OF REAL;
BEGIN (*RungeKutta4Proc*)
  n:= HIGH(x);
  IF n>nmax-1 THEN (*too big model order*) 
    Error(badModelOrder)
  END;
  hh:= h/2.0;
  
  CurrentEquationProcedure(x,dx);			(*1st evaluation*)
  FOR i:= 0 TO n DO x1[i]:= x[i]+hh*dx[i] END;
  time:= time + hh;
  
  CurrentEquationProcedure(x1,dx1);			(*2nd evaluation*)
  FOR i:= 0 TO n DO x2[i]:= x[i]+hh*dx1[i] END;
  
  CurrentEquationProcedure(x2,dx2);			(*3rd evaluation*)
  FOR i:= 0 TO n DO x3[i]:= x[i]+h*dx2[i] END;
  time:= time + hh;
  
  CurrentEquationProcedure(x3,dx3);			(*4th evaluation*)
  FOR i:= 0 TO n DO 
    xnew[i]:= x[i]+h/6.0*(dx[i]+2.0*dx1[i]+2.0*dx2[i]+dx3[i])
  END;
END RungeKutta4Proc;


PROCEDURE DiscreteTimeProc(VAR x,xnew: ARRAY OF REAL);
VAR
BEGIN (*DiscreteTimeProc*)
  CurrentEquationProcedure(x,xnew);
  time:= time+1.0; (* ist das eine gute Idee?*)
END DiscreteTimeProc;
 	

PROCEDURE SetIntegrationMethod(m:IntegrationMethod);
BEGIN (*SetIntegrationMethod*)
  CASE m OF
    Euler		: Integrate:= EulerProc;
    			  currentMethod:= m;
  | Heun		: Integrate:= HeunProc;
    			  currentMethod:= m;
  | RungeKutta4		: Integrate:= RungeKutta4Proc;
    			  currentMethod:= m;
  | DiscreteTime	: Integrate:= DiscreteTimeProc;
    			  currentMethod:= m;
  ELSE
    Error(methodNotImplemented);
  END; (*CASE*)
END SetIntegrationMethod;


PROCEDURE GetIntegrationMethod(VAR m:IntegrationMethod);
BEGIN
  m:= currentMethod;
END GetIntegrationMethod;


PROCEDURE SetIntegrationStep(step: REAL);
BEGIN
  h:= step;
END SetIntegrationStep;


PROCEDURE GetIntegrationStep(VAR step: REAL);
BEGIN
  step:= h;
END GetIntegrationStep;


PROCEDURE NoModelWarning(VAR x,y: ARRAY OF REAL);
  VAR i: INTEGER;
BEGIN   
  INC(errorCount);
  IF errorCount<3 THEN Error(noModelInstalled) END;
  FOR i:= 0 TO HIGH(x) DO x[i]:= 0.0 END;
  FOR i:= 0 TO HIGH(y) DO y[i]:= 0.0 END;
END NoModelWarning;

 
BEGIN (*SimLib*)
    SetIntegrationMethod(Heun);
    SetIntegrationStep(0.05);
    errorCount:= 0;
    InstallModel(NoModelWarning);
END SimLib0.