
  IMPLEMENTATION MODULE SortLib;

  (******************************************************************)
  (*                                                                *)
  (*        MODULE: SortLib,              Version   0.1 / 001       *)
  (*                                                                *)
  (*        Purpose:                                                *)
  (*                                                                *)
  (*        Implementation:  For Apple Macintosh                    *)
  (*                                                                *)
  (*        Programming:     O. Roth                                *)
  (*                         Institut fr Phytomedizin              *)
  (*                         ETH-Zentrum /CLS                       *)
  (*                         CH-8092 Zrich                         *)
  (*                                                                *)
  (*        LAST MODIFICATION :   2. 8.87                           *)
  (******************************************************************)


  FROM SYSTEM IMPORT VAL;



  PROCEDURE QuickSortX ( VAR a: ARRAY OF REAL; n: CARDINAL );
  (***********************************************************)
  (*  Sorts an array a by the method of C.A.R. HOARE.        *)                      (*  Fast for large n.                                      *)
  (*  Algorithm from:                                        *)
  (*  Niklaus Wirth's " Algoritmen und Datenstrukturen,      *)
  (*  B.G. Teubner Stuttgart, 1983.                          *)
  (***********************************************************)


    CONST
      m = 16;

    VAR
      i, j, l, r: INTEGER;
      x, w:       REAL;
      s:          [0..m];
      stack:      ARRAY[1..m] OF RECORD
                                   l, r: CARDINAL;
                                 END;

  BEGIN  (* QuickSortX *)

    s:= 1;
    stack[1].l:= 0;
    stack[1].r:= n-1;

    REPEAT                          (* nimm oberste Anweisung vom Stack *)
      l:= stack[s].l;
      r:= stack[s].r;
      DEC( s );

      REPEAT                        (* teile  a[l], ... , a[r]  auf *)
        i:= l;
        j:= r;
        x:= a[ (l+r) DIV 2 ];

        REPEAT
          WHILE ( a[i] < x    )  DO  INC( i );  END; (* WHILE *)
          WHILE (    x < a[j] )  DO  DEC( j );  END; (* WHILE *)
          IF ( i <= j ) THEN
            w:= a[i];
            a[i]:= a[j];
            a[j]:= w;
            INC( i );
            DEC( j );
          END; (* IF *)
        UNTIL ( i > j );
                                    (* speichere Zerlegungs-Anweisung:  *)
        IF ( (j-l) < (r-i) ) THEN   (* Stack-Eintragungen fr Sortieren *)
          IF ( i < r ) THEN         (* des rechten Teils *)
            INC( s );
            stack[s].l:= i;
            stack[s].r:= r;
          END; (* IF *)
          r:= j;                    (* Weitersortieren des linken Teils *)
        ELSE
          IF ( l < j ) THEN         (* Stack-Eintragungen fr Sortieren *)
            INC( s );               (* des linken Teils *)
            stack[s].l:= l;
            stack[s].r:= j;
          END; (* IF *)
          l:= i;                    (* Weitersortieren des rechten Teils *)
        END; (* IF *)
        
      UNTIL ( l >= r );
        
    UNTIL ( s = 0 );
      
  END QuickSortX ;




  PROCEDURE QuickSortXY ( VAR a, b: ARRAY OF REAL; n: CARDINAL );
  (***********************************************************)
  (*  Sorts an array a and at the same time the parallel     *)
  (*  array b by the method of C.A.R. HOARE.                 *)                      (*  Fast for large n.                                      *)
  (*  Algorithm from:                                        *)
  (*  Niklaus Wirth's " Algoritmen und Datenstrukturen,      *)
  (*  B.G. Teubner Stuttgart, 1983.                          *)
  (***********************************************************)

    CONST
      m = 16;

    VAR
      i, j, l, r: INTEGER;
      x, w, t:    REAL;
      s:          [0..m];
      stack:      ARRAY[1..m] OF RECORD
                                   l, r: CARDINAL;
                                 END;

  BEGIN  (* QuickSortXY *)

    s:= 1;
    stack[1].l:= 0;
    stack[1].r:= n-1;

    REPEAT                          (* nimm oberste Anweisung vom Stack *)
      l:= stack[s].l;
      r:= stack[s].r;
      DEC( s );

      REPEAT                        (* teile  a[l], ... , a[r]  auf *)
        i:= l;
        j:= r;
        x:= a[ (l+r) DIV 2 ];

        REPEAT
          WHILE ( a[i] < x    )  DO  INC( i )  END; (* WHILE *)
          WHILE (    x < a[j] )  DO  DEC( j )  END; (* WHILE *)
          IF ( i<= j ) THEN
            w:= a[i];
            a[i]:= a[j];
            a[j]:= w;
            t:= b[i];
            b[i]:= b[j];
            b[j]:= t;
            INC( i );
            DEC( j );
          END; (* IF *)

        UNTIL ( i > j );
                                    (* speichere Zerlegungs-Anweisung:  *)
        IF ( (j-l) < (r-i) ) THEN   (* Stack-Eintragungen fr Sortieren *)
          IF ( i < r ) THEN         (* des rechten Teils *)
            INC( s );
            stack[s].l:= i;
            stack[s].r:= r;
          END; (* IF *)
          r:= j;                    (* Weitersortieren des linken Teils *)
        ELSE
          IF ( l < j ) THEN         (* Stack-Eintragungen fr Sortieren *)
            INC( s );               (* des linken Teils *)
            stack[s].l:= l;
            stack[s].r:= j;
          END; (* IF *)
          l:= i;                    (* Weitersortieren des rechten Teils *)
        END; (* IF *)
        
      UNTIL ( l >= r );
        
    UNTIL ( s = 0 );
      
  END QuickSortXY ;




  PROCEDURE BinarySortX ( VAR a: ARRAY OF REAL; n: CARDINAL );
  (***********************************************************)
  (*  Sorts an array a by the method of binary insertion.    *)                      (*  Fast for small n.                                      *)
  (*  Algorithm from:                                        *)
  (*  Niklaus Wirth's " Algoritmen und Datenstrukturen,      *)
  (*  B.G. Teubner Stuttgart, 1983.                          *)
  (***********************************************************)

    VAR
      i, j, l, r, m, N: INTEGER;
      x:             REAL;

  BEGIN (* BinarySortX *)

    N:= VAL( INTEGER, n ) -1;
    FOR i:= 1 TO N DO
      x:= a[i];
      l:= 0;
      r:= i-1;
      WHILE ( l <= r ) DO
        m:= (l+r) DIV 2;
        IF ( x < a[m] ) THEN  r:= m-1;
        ELSE                  l:= m+1;
        END; (* IF *)
      END; (* WHILE *)
      FOR j:= i-1 TO l BY -1 DO a[j+1]:= a[j]; END; (* FOR *)
      a[l]:= x;
    END; (* FOR *)
    
  END BinarySortX ;




  PROCEDURE StrSelSortX ( VAR a: ARRAY OF REAL; n: CARDINAL );
  (***********************************************************)
  (*  Sorts an array a by the method of straight selection.  *)                      (*  Relatively fast for small n.                           *)
  (*  Algorithm from:                                        *)
  (*  Niklaus Wirth's " Algoritmen und Datenstrukturen,      *)
  (*  B.G. Teubner Stuttgart, 1983.                          *)
  (***********************************************************)
  
    VAR
      i, j, k:  INTEGER;
      nm2, nm1: INTEGER;
      x:        REAL;

  BEGIN (* StrSelSortX *)

    nm2:= n-2;
    nm1:= n-1;
    FOR i := 0 TO nm2 DO
      k:= i;
      x:= a[i];
      FOR j := i+1 TO nm1 DO
        IF ( a[j] < x ) THEN
          k:= j;
          x:= a[j];
        END; (* IF *)
      END; (* FOR *)
      a[k]:= a[i];
      a[i]:= x;
    END; (* FOR *)
    
  END StrSelSortX ;




  PROCEDURE StrSelSortXY ( VAR a, b: ARRAY OF REAL; n: CARDINAL );
  (***********************************************************)
  (*  Sorts an array a and at the same time the parallel     *)
  (*  array b by the method of straight selection.           *)                      (*  Fast for small n.                                      *)
  (*  Algorithm from:                                        *)
  (*  Niklaus Wirth's " Algoritmen und Datenstrukturen,      *)
  (*  B.G. Teubner Stuttgart, 1983.                          *)
  (***********************************************************)
  
    VAR
      i, j, k:  INTEGER;
      nm1, nm2: INTEGER;
      x, y:     REAL;

  BEGIN (* StrSelSortXY *)

    nm1:= n-1;
    nm2:= n-2;
    FOR i := 0 TO nm2 DO
      k:= i;
      x:= a[i];
      y:= b[i];
      FOR j := i+1 TO nm1 DO
        IF ( a[j] < x ) THEN
          k:= j;
          x:= a[j];
          y:= b[j];
        END; (* IF *)
      END; (* FOR *)
      a[k]:= a[i];
      a[i]:= x;
      b[k]:= b[i];
      b[i]:= y;
    END; (* FOR *)
    
  END StrSelSortXY ;



END SortLib.
