
IMPLEMENTATION MODULE StatLib;

  (******************************************************************)
  (*                                                                *)
  (*        MODULE: StatLib               Version   0.1 / 001       *)
  (*                                                                *)
  (*        Purpose:                                                *)
  (*                                                                *)
  (*        Implementation:  For Apple Macintosh                    *)
  (*                                                                *)
  (*        Programming:     O. Roth                                *)
  (*                         Institut fr Phytomedizin              *)
  (*                         ETH-Zentrum /CLS                       *)
  (*                         CH-8092 Zrich                         *)
  (*                                                                *)
  (*        LAST MODIFICATION :   09/01/91  dg                      *)
  (******************************************************************)


  FROM DMMathLib IMPORT Exp, Sqrt;


PROCEDURE MinX ( VAR X: ARRAY OF REAL;  N: CARDINAL ): REAL;
(* MinX returns the minimum value of the first N reals of array X. *)
  VAR
    minx : REAL;
    i    : CARDINAL;
BEGIN
  minx:= X[0];
  FOR i:= 1 TO N-1 DO
    IF minx > X[i] THEN  minx:= X[i];  END;
  END; (* FOR *)
  RETURN minx;
END MinX;



PROCEDURE MaxX ( VAR X: ARRAY OF REAL;  N: CARDINAL ): REAL;
(* MaxX returns the maximum value of the first N reals of array X. *)
  VAR
    maxx : REAL;
    i    : CARDINAL;
BEGIN
  maxx:= X[0];
  FOR i:= 1 TO N-1 DO
    IF maxx < X[i] THEN  maxx:= X[i];  END;
  END; (* FOR *)
  RETURN maxx;
END MaxX;


PROCEDURE WSumX ( VAR X: ARRAY OF REAL;
                  N    : CARDINAL;
                  FX   : FunctionXProc ): REAL;
(* WSumX returns the weighted sum of the first N reals of array X.    *)
(* Each value is weighted according to the function FX. For instance, *)
(* to compute the sum of the squares of the first 10 values of X, you *)
(* should define a function such as Square( x: REAL ): REAL; , which  *)
(* returns the square of x. Once this is done, WSumX( X, 10, Square );*)
(* would return the sum of squares of the first 10 values of X.       *)
  VAR
    sumx : REAL;
    i    : CARDINAL;
BEGIN
  sumx:= 0.0;
  FOR i:= 0 TO N-1 DO
    sumx:= sumx + FX( X[i] );
  END; (* FOR *)
  RETURN sumx;
END WSumX;


PROCEDURE SumX ( VAR X: ARRAY OF REAL;  N: CARDINAL ): REAL;
(* SumX returns the sum of X[i] for the first N values of X. *)
  VAR
    sumx : REAL;
    i    : CARDINAL;
BEGIN
  sumx:= 0.0;
  FOR i:= 0 TO N-1 DO
    sumx:= sumx + X[i];
  END; (* FOR *)
  RETURN sumx;
END SumX;


PROCEDURE SumXY ( VAR X, Y: ARRAY OF REAL;  N: CARDINAL ): REAL;
(* SumXY returns the sum of X[i]*Y[i] for the first N values of X and Y.*)
  VAR
    sumxy : REAL;
    i    : CARDINAL;
BEGIN
  sumxy:= 0.0;
  FOR i:= 0 TO N-1 DO
    sumxy:= sumxy + X[i]*Y[i];
  END; (* FOR *)
  RETURN sumxy;
END SumXY;


PROCEDURE SumX2 ( VAR X: ARRAY OF REAL;  N: CARDINAL ): REAL;
(* SumX2 returns the sum of X[i]^2 for the first N values of X. This *)
(* issomewhat faster than using WSumX as shown in the example above. *)
  VAR
    sumx2 : REAL;
    i    : CARDINAL;
BEGIN
  sumx2:= 0.0;
  FOR i:= 0 TO N-1 DO
    sumx2:= sumx2 + X[i]*X[i];
  END; (* FOR *)
  RETURN sumx2;
END SumX2;


PROCEDURE SumX3 ( VAR X: ARRAY OF REAL;  N: CARDINAL ): REAL;
(* SumX3 returns the sum of X[i]^3 for the first N values of X. *)
  VAR
    sumx3 : REAL;
    i    : CARDINAL;
BEGIN
  sumx3:= 0.0;
  FOR i:= 0 TO N-1 DO
    sumx3:= sumx3 + X[i]*X[i]*X[i];
  END; (* FOR *)
  RETURN sumx3;
END SumX3;


PROCEDURE SumX4 ( VAR X: ARRAY OF REAL;  N: CARDINAL ): REAL;
(* SumX4 returns the sum of X[i]^4 for the first N values of X. *)
  VAR
    sumx4 : REAL;
    i    : CARDINAL;
BEGIN
  sumx4:= 0.0;
  FOR i:= 0 TO N-1 DO
    sumx4:= sumx4 + X[i]*X[i]*X[i]*X[i];
  END; (* FOR *)
  RETURN sumx4;
END SumX4;


PROCEDURE WMeanX ( VAR X: ARRAY OF REAL;
                   N    : CARDINAL;
                   FX   : FunctionXProc ): REAL;
(* WMeanX returns the weighted mean of the first N reals of array X. *)
(* Each X[i] is weighted according to the function FX.               *)
BEGIN
  RETURN ( WSumX( X, N, FX ) / FLOAT(N) );
END WMeanX;


PROCEDURE MeanX ( VAR X: ARRAY OF REAL;  N: CARDINAL ): REAL;
(* MeanX returns the mean of the first N reals of array X. *)
BEGIN
  RETURN ( SumX( X, N ) / FLOAT(N) );
END MeanX;


PROCEDURE VarX ( VAR X: ARRAY OF REAL;  N: CARDINAL ): REAL;
(* VarX returns the variance of the first N reals of array X. *)
  VAR
    sdev : REAL;
BEGIN
  sdev:= SDevX( X, N );
  RETURN ( sdev*sdev );
END VarX;


PROCEDURE SDevX ( VAR X: ARRAY OF REAL;  N: CARDINAL ): REAL;
(* SDevX returns the standard deviation of the first N reals of array X.*)
  VAR
    sumx, sumx2 : REAL;
BEGIN
  sumx := SumX ( X, N );
  sumx2:= SumX2( X, N );
  RETURN ( Sqrt( FLOAT(N)*sumx2 - sumx*sumx) / FLOAT(N) );
END SDevX;


PROCEDURE SkewX ( VAR X: ARRAY OF REAL;  N: CARDINAL ): REAL;
(* SkewX returns the skewness of the first N reals of array X.*)
  VAR
    l, sumx, sumx2, sumx3, rN : REAL;
BEGIN
  sumx := SumX ( X, N );
  sumx2:= SumX2( X, N );
  sumx3:= SumX3( X, N );
  rN:= FLOAT(N);
  l:= SDevX( X, N ) * rN;
  RETURN ( (rN*rN*sumx3 -3.0*rN*sumx*sumx2 +2.0*sumx*sumx) / (l*l*l) );
END SkewX;


PROCEDURE KurtX ( VAR X: ARRAY OF REAL;  N: CARDINAL ): REAL;
(* KurtX returns the kurtosis of the first N reals of array X.*)
  VAR
    l, sumx, sumx2, sumx3, sumx4, rN : REAL;
BEGIN
  sumx := SumX ( X, N );
  sumx2:= SumX2( X, N );
  sumx3:= SumX3( X, N );
  sumx4:= SumX4( X, N );
  rN:= FLOAT(N);
  l:= SDevX( X, N ) * rN;
  RETURN ( ( rN*rN*rN*sumx4 
             -4.0*rN*rN*sumx*sumx3
             +6.0*rN*sumx*sumx*sumx2
             -3.0*sumx*sumx*sumx*sumx) /  ( l*l*l*l ) );
END KurtX;


PROCEDURE LinearReg ( VAR X, Y     : ARRAY OF REAL;
                      N            : CARDINAL;
                      VAR a, b, r2 : REAL         );
(* LinearReg performs linear regression analysis on the first N    *)
(* values of X and Y. a and b are the coefficients of the best fit *)
(* straight line, Y= ax+b; r2 is the coefficient of determination  *)
(* approaches zero as the data fail to fit a straight line.        *)
  VAR
    d, sumx, sumx2, sumy, sumy2, sumxy, rN : REAL;
BEGIN
  sumx := SumX ( X, N );
  sumx2:= SumX2( X, N );
  sumy := SumX ( Y, N );
  sumy2:= SumX2( Y, N );
  sumxy:= SumXY( X, Y, N );
  rN:= FLOAT(N);
  d:= rN*sumx2 - sumx*sumx;
  a:= (sumy*sumx2 - sumxy*sumx) / d;
  b:= (rN*sumxy - sumx*sumy) / d;
  r2:= (a*sumy + b*sumxy - sumy*sumy/rN) / (sumy2 - sumy*sumy/rN);
END LinearReg;


PROCEDURE FuncX ( VAR X, Y: ARRAY OF REAL;
                  N       : CARDINAL;
                  FX      : FunctionXProc );
(* FuncX takes the first N values of array X and computes   *)
(* the corresponding Y values such that Y[i]= FX( X[i] );.  *)
  VAR
    i: CARDINAL;
BEGIN
  FOR i:= 0 TO N-1 DO
    Y[i]:= FX( X[i] );
  END; (* FOR *)
END FuncX;


PROCEDURE CountX ( VAR X      : ARRAY OF REAL;
                   N          : CARDINAL;
                   xLow, xHigh: REAL;
                   InRangeX   : InRangeProc  ): CARDINAL;
(* CountX returns the number of X values (X[0]..X[N-1]) that meet    *)
(* the criteria defined by the boolean function InRangeX. See the    *)
(* description of InRangeProc above. Although XLow and XHigh are     *)
(* commonly used to define a continuous range, this is not necessary.*)
(* For instance, InRangeX( x, xLow, xHigh ) may be written to yield  *)
(* TRUE if x is greater than XHigh or less than XLow.                *)
  VAR
    i, count: CARDINAL;
BEGIN
  count:= 0;
  FOR i:= 0 TO N-1 DO
    IF InRangeX( X[i], xLow, xHigh ) THEN  INC( count );  END; (* IF *)
  END; (* FOR *)
  RETURN count;
END CountX;


PROCEDURE InsertX ( VAR X  : ARRAY OF REAL;
                    N, j   : CARDINAL;
                    xValue : REAL       );
(* InsertX is an array manipulation procedure. It first shifts     *)
(* X[j]..X[N-1] one index higher. xValue is then assigned to X[j]. *)
(* Notice that the overall effect is to insert xValue into X[j].   *)
(* Remember that the lower limit of open arrays is zero, and that  *)
(* the value parameter N is not updated (i.e. incremented by 1).   *) 
  VAR
    i, NN, hn : CARDINAL;
BEGIN
  hn:= HIGH( X );
  NN:= N;
  IF ( N > hn ) THEN  NN:= HIGH(X);  END;
  IF ( j < hn ) THEN
    FOR i:= NN TO j+1 BY -1 DO
      X[i]:= X[i-1];
    END; (* FOR *)
  END; (* IF *)
  X[j]:= xValue;
END InsertX;


PROCEDURE DeleteX ( VAR X  : ARRAY OF REAL;
                    N, j   : CARDINAL  );
(* DeleteX is an array manipulation procedure. It deletes X[j] by  *)
(* assigning X[i]:= X[N+1], for i:=j to N-1. This is the opposite  *)
(* of InsertX.                                                     *)
  VAR
    i, hn : CARDINAL;
BEGIN
  hn:= HIGH( X );
  IF ( j < hn ) THEN
    FOR i:= j+1 TO N-1 DO
      X[i-1]:= X[i];
    END; (* FOR *)
  END; (* IF *)
END DeleteX;


PROCEDURE ClearX ( VAR X  : ARRAY OF REAL;
                   N      : CARDINAL;
                   xValue : REAL       );
(* ClearX is an array manipulation procedure. It assigns the xValue *)
(* to the first N values of X.                                      *)
  VAR
    i : CARDINAL;
BEGIN
  FOR i:= 0 TO N-1 DO
    X[i]:= xValue;
  END; (* FOR *)
END ClearX;


PROCEDURE SortX ( VAR X: ARRAY OF REAL;  N: CARDINAL );
(* SortX sorts the first N reals of array X in ascending order. This *)
(* is an implementation of the QuickSort algorithm.                  *)

  PROCEDURE sort ( l, r: CARDINAL );
    VAR
      i, j    : CARDINAL;
      w, x, y : REAL;
  BEGIN
    i:= l;
    j:= r;
    x:= X[(l+r) DIV 2];
    REPEAT
      WHILE (X[i] < x) DO   INC( i );   END;
      WHILE (X[j] > x) DO   DEC( j );   END;
      IF ( i<= j ) THEN (* swap *)
        w   := X[i];
        X[i]:= X[j];
        X[j]:= w;
        INC( i );
        DEC( j );
      END; (* IF *)
    UNTIL ( i > j );
    IF ( l < j ) THEN  sort( l, j );  END;
    IF ( i < r ) THEN  sort( i, r );  END;
  END sort;
  
BEGIN
  sort( 0, N-1 );
END SortX;


PROCEDURE NormDist ( z1, z2 : REAL ): REAL;
(* NormDist returns the area under the normal distribution curve whith  *)
(* a lower limit of z1 and an upper limit of z2. Note that NormDist     *)
(* returns 1.0 as z1 approaches negative infinity and z2 approaches     *)
(* positive infinity. If X is a random variable, then one would expect  *)
(* approxiamtely 68 percent of the observed values of X to fall whithin *)
(* plus or minus one standard deviation of the mean of X.               *)
  CONST
    maxiterations = 30;
  VAR
    i : INTEGER;
    x, h, a, s : REAL;
  PROCEDURE F ( x: REAL ): REAL;
  BEGIN
    RETURN Exp( -x*x/2.0 );
  END F;
  
BEGIN
  h:= (z2-z1)/FLOAT(maxiterations);
  a:= F( z1 ) + F( z2 );
  s:= 0.0;
  FOR i:= 1 TO maxiterations-1 BY 2 DO
    s:= s + F( z1 + FLOAT(i) * h );
  END; (* FOR *)
  a:= a+4.0*s;
  s:= 0.0;
  FOR i:= 2 TO maxiterations-2 BY 2 DO
    s:= s + F( z1 + FLOAT(i) * h );
  END; (* FOR *)
  a:= (h/3.0)*(a+s*2.0);
  RETURN ( a * 0.398942 );
END NormDist;


PROCEDURE Factorial ( N: CARDINAL ): REAL;
(* Factorial returns N! (N factorial).  *)
  VAR
    i    : CARDINAL;
    fact : REAL;
BEGIN
  fact:= 1.0;
  IF N=0 THEN
    fact:= 1.0;
  ELSE
    FOR i:= 1 TO N DO
      fact:= fact*FLOAT(i);
    END; (* FOR *)
  END; (* IF *)
  RETURN fact;
END Factorial;


PROCEDURE Combination ( N, R : CARDINAL ): REAL;
(* Combination returns the number of combinations of N objects *)
(* taken R at a time.                                          *)
BEGIN
  RETURN ( Permutation( N, R ) / Factorial( R ) );
END Combination;

PROCEDURE Permutation ( N, R : CARDINAL ): REAL;
(* Permutation returns the number of permutations of N objects *)
(* taken R at a time.                                          *)
  VAR
    p: REAL;
    i: CARDINAL;
BEGIN
  p:= 1.0;
  FOR i:= N-R+1 TO N DO
    p:= p*FLOAT(i);
  END; (* FOR *)
  RETURN p;
END Permutation;


BEGIN
END StatLib.
  